unit U_PreLoad;
interface
uses
  Windows, Messages, SysUtils, Classes,
  {$ifndef NODB} Db, {$endif}
  ZLib;

// File format:

type
  TPlf_Header=record
    Signature: array[0..7] of char; // ZLIB_PLF
    Version:   LongInt;
    DirOffset: LongInt;
    Reserved:  LongInt;
  end;

  TPlf_DirHeader=record
    Signature: array[0..7] of char; // ZLIB_DIR 
    DirSize:   LongInt;
    NumItems:  LongInt;
    Reserved:  LongInt;
  end;

  TDateRec=packed record
    year:  ShortInt; // Signed byte, +2000   (1873-2127)
    month: byte;
    day:   byte;
    fill:  byte;
  end;

  TPlf_DirItem=record
    ItemSize:   LongInt;  // =SizeOf(TPlf_DirItem) - 4 !    (excluding ItemSize)
    DataOffset: LongInt;
    DataSize:   LongInt;
    CleanSize:  LongInt;
    SpaceAfter: LongInt;  // initially 0-3, when replaced, it may be larger...
    Datum:      TDateRec;
    Reserved:   LongInt;
    Name:       string[15];
  end;
  PPlf_DirItem=^TPlf_DirItem;

const
  sPlf_Signature:PChar='ZLIB_PLF';
  sPlf_Dir_Signature:PChar='ZLIB_DIR';
  cPlf_Version=$173FE513;
  cPlf_Extension='.PLF';

// Object interface:
type
  TPreLoadObject=class(TObject)
  private
    function GetItemCount: integer;
    function GetItemName(Index: integer): string;
    function GetItemDate(Index: integer): TDateTime;
    function GetItemSize(Index: integer): integer;
  protected
    FFileName:  string;
    FStream:    TFileStream;
    FDirOffset: integer;
    FDirRes:    integer;
    FDir:       TStringList;
    procedure   ReadDirectory;
    procedure   ClearDirectory;
  public
    constructor CreatePreLoad(const FileName: string; WriteMode: Boolean);
    destructor  Destroy; override;
    property  FileName: string read FFileName;
    property  ItemCount: integer read GetItemCount;
    property  Items[Index: integer]: string read GetItemName;
    property  ItemDate[Index: integer]: TDateTime read GetItemDate;
    property  ItemSize[Index: integer]: integer read GetItemSize; // CleanSize...
    function  IndexOfItem(const Name: string): integer;
  end;

  // Reader and Writer are separated, because the later would be eliminated by linker in client apps...
  TPreLoadReader=class(TPreLoadObject)
  protected
  public
    constructor Create(const FileName: string); virtual; // File must exist!
    function OpenItem(Index: integer; out Reader: TStream; out Size: integer): Boolean; // Returns non-seakable stream!
  end;

  TPreLoadWriter=class(TPreLoadReader)
  private
    procedure InternalStoreItem(const Name: string; Data: TStream; IsCompressed: Boolean; CleanSize: integer; const Date: TDateTime);
  protected
    DirClean: Boolean;
    procedure ReplaceItem(Index: integer; Data: TStream; const Date: TDateTime);
  public
    constructor Create(const FileName: string); override; // File is created, if it does not exist!
    destructor Destroy; override;
    procedure StoreItem(const Name: string; Data: TStream; const Date: TDateTime);
    procedure WriteDirectory; // called from destructor...
    procedure DeleteItem(Index: integer);
    procedure ImportArchive(Src: TPreLoadObject);
  end;

{$ifndef NODB}
type
  TDbPreLoadReader=class(TPreLoadReader)
  protected
  public
    function  HasTable(const TableName: string): Boolean;
    function  LoadTable(const TableName: string; Tbl: TDataset): Boolean;
  end;

  TDbPreLoadWriter=class(TPreLoadWriter)
  private
    procedure StoreTableEx(const TableName: string; Tbl: TDataset; Fields: TStrings);
  protected
  public
    function  HasTable(const TableName: string): Boolean;
    function  LoadTable(const TableName: string; Tbl: TDataset): Boolean;
    procedure StoreTable(const TableName: string; Tbl: TDataset);
  end;

function NormalizeTableName(const TableName: string): string;

{$endif NOT NODB}

implementation
uses ComUtil
     {$ifndef NODB},Base64 {$endif NOT NODB};

resourcestring
  SzBadVersion='patn verze inicializanho souboru!';
  SzPlfDamaged='Inicializan soubor je pokozen!';

{ TPreLoadObject }

constructor TPreLoadObject.CreatePreLoad(const FileName: string; WriteMode: Boolean);
var Mode: integer;
    NewFile: Boolean;
    Hdr: TPlf_Header;
    Dir: TPlf_DirHeader;
begin
  inherited Create;
  Self.FFileName:=ExpandFileName(FileName);
  NewFile:=False;
  if WriteMode then begin
    if FileExists(FileName) then begin
      Mode:=fmOpenReadWrite or fmShareDenyWrite;
    end else begin
      Mode:=fmCreate;
      NewFile:=True;
    end;
  end else begin
    Mode:=fmOpenRead or fmShareDenyWrite;
  end;
  FDir:=TStringList.Create;
  FStream:=TFileStream.Create(FileName,Mode);
  if NewFile then begin
    // File header:
    FillChar(Hdr,SizeOf(Hdr),0);
    Move(sPlf_Signature^,Hdr.Signature[0],SizeOf(Hdr.Signature));
    Hdr.Version:=cPlf_Version;
    Hdr.DirOffset:=sizeof(Hdr);
    FStream.WriteBuffer(Hdr,SizeOf(Hdr));
    // Dir header:
    FDirOffset:=FStream.Position;
    Move(sPlf_Dir_Signature^,Dir.Signature[0],SizeOf(Dir.Signature));
    FillChar(Dir,SizeOf(Dir),0);
    Dir.DirSize:=0;
    Dir.NumItems:=0;
    FStream.WriteBuffer(Dir,SizeOf(Dir));
  end else begin
    FStream.ReadBuffer(Hdr,SizeOf(Hdr));
    // Version verification:   (would change in newer versions:)
    if Hdr.Version<>cPlf_Version then
      raise Exception.Create(SzBadVersion);
    //
    FDirOffset:=Hdr.DirOffset;
    ReadDirectory;
  end;
end;

destructor TPreLoadObject.Destroy;
begin
  SaveFree(@FStream);
  ClearDirectory;
  SaveFree(@FDir);
  inherited;
end;

procedure TPreLoadObject.ClearDirectory;
var i: integer;
    item: PPlf_DirItem;
begin
  if FDir=nil then exit;
  for i:=FDir.Count-1 downto 0 do begin
    item:=Pointer(FDir.Objects[i]);
    FDir.Delete(i);
    if item<>nil then
      FreeMem(item);
  end;
end;

procedure TPreLoadObject.ReadDirectory;
var Dir: TPlf_DirHeader;
    i: integer;
    Size: LongInt;
    item: PPlf_DirItem;
begin
  FStream.Position:=FDirOffset;
  FStream.ReadBuffer(Dir,SizeOf(Dir));
  FDirRes:=Dir.Reserved; // save reserved value... (it may be of some importance???)
  for i:=0 to Dir.NumItems-1 do begin
    FStream.ReadBuffer(Size,SizeOf(LongInt));
    if (Size+SizeOf(LongInt)<SizeOf(TPlf_DirItem)) or (Size>=16*1024) then
      raise Exception.Create(SzPlfDamaged);
    item:=AllocMem(Size+SizeOf(LongInt));
    try
      item.ItemSize:=Size;
      FStream.Read((PChar(item)+SizeOf(LongInt))^,Size);
      FDir.AddObject(item.Name,TObject(item));
    except
      FreeMem(item);
      raise;
    end;
  end;
end;

function TPreLoadObject.GetItemCount: integer;
begin
  Result:=FDir.Count;
end;

function TPreLoadObject.GetItemName(Index: integer): string;
begin
  Result:=FDir[Index];
end;

function TPreLoadObject.IndexOfItem(const Name: string): integer;
var i: integer;
begin
  for i:=0 to FDir.Count-1 do
    if AnsiCompareText(Name,FDir[i])=0 then begin
      Result:=i;
      exit;
    end;
  Result:=-1;
end;

function UnPackDate(const Datum: TDateRec): TDateTime;
begin
  Result:=0;
  if (Datum.month>0) and (Datum.month<=12) and (Datum.day>0) and (Datum.day<=31) then
    try
      Result:=EncodeDate(Datum.year+2000,Datum.month,Datum.day);
    except ; end;
end;

function TPreLoadObject.GetItemDate(Index: integer): TDateTime;
var item: PPlf_DirItem;
begin
  item:=Pointer(FDir.Objects[Index]);
  Result:=UnPackDate(item.Datum);
end;

function TPreLoadObject.GetItemSize(Index: integer): integer;
var item: PPlf_DirItem;
begin
  item:=Pointer(FDir.Objects[Index]);
  Result:=Item.CleanSize;
end;

{ TPreLoadReader }

type
  TOwnDecompressor=class(TDecompressionStream)
  public
    OwnedStream: TStream;
    destructor Destroy; override;
  end;

{ TOwnDecompressor }

destructor TOwnDecompressor.Destroy;
begin
  try
    inherited;
  finally SaveFree(@OwnedStream); end;
end;

constructor TPreLoadReader.Create(const FileName: string);
begin
  inherited CreatePreLoad(FileName,False);  
end;

function TPreLoadReader.OpenItem(Index: integer; out Reader: TStream; out Size: integer): Boolean;
var item: PPlf_DirItem;
    Sub: TSubStream;
    Deco: TOwnDecompressor;
begin
  Reader:=nil;
  Size:=0;
  Result:=False;
  if (Index<0) or (Index>=FDir.Count) then exit;
  item:=Pointer(FDir.Objects[Index]);
  if (item.DataOffset<sizeof(TPlf_Header)) or (item.DataOffset+item.DataSize>FDirOffset)
     or (item.DataSize<0)
  then
    exit;
  if item.DataSize=0 then begin
    Reader:=TMemoryStream.Create; // return empty stream, Size=0...
    Result:=True;
    exit;
  end;
  //FStream.Position:=item.DataOffset;
  Sub:=TSubStream.Create(FStream,item.DataOffset,item.DataSize);
  Sub.ReadOnly:=True;
  Deco:=TOwnDecompressor.Create(Sub);
  Deco.OwnedStream:=Sub;
  Reader:=Deco;
  Size:=item.CleanSize;
  Result:=True;
end;

{ TPreLoadWriter }

constructor TPreLoadWriter.Create(const FileName: string);
begin
  DirClean:=True;
  inherited CreatePreLoad(FileName,True);
end;

procedure NotImplemented;
begin
  raise Exception.Create('Not yet implemented!!!');
end;

destructor TPreLoadWriter.Destroy;
begin
  if not DirClean then WriteDirectory;
  inherited;
end;

procedure Compress(InStream,OutStream: TStream; out CleanSize: integer);
var buf: array[0..4095] of char;
    sz:  integer;
    Comp: TCompressionStream;
begin
  CleanSize:=0;
  Comp:=TCompressionStream.Create(clMax,OutStream);
  try
    while True do begin
      sz:=InStream.Read(buf[0],SizeOf(buf));
      if sz<=0 then break;
      inc(CleanSize,sz);
      Comp.WriteBuffer(buf[0],sz);
    end;
  finally Comp.Free; end;
end;

function PackDate(const Date: TDateTime): TDateRec;
var y,m,d: Word;
begin
  if Date=0 then
    DecodeDate(SysUtils.Date,y,m,d)
  else
    DecodeDate(Date,y,m,d);
  Result.year:=y-2000;
  Result.month:=m;
  Result.day:=d;
  Result.fill:=0;
end;

procedure TPreLoadWriter.ReplaceItem(Index: integer; Data: TStream; const Date: TDateTime);
var ms: TMemoryStream;
    item: PPlf_DirItem;
    NewSize,CleanSize: integer;
    Name: string;
    buf: array[0..31] of byte;
    i: integer;
begin
  item:=Pointer(FDir.Objects[Index]);
  ms:=TMemoryStream.Create;
  try
    // Compress into TMemoryStream to see, if it fits...
    Data.Position:=0;
    Compress(Data,ms,CleanSize);
    NewSize:=ms.Size;
    if NewSize<=item.DataSize+item.SpaceAfter then begin
      // Replace the item in-place:
      item.CleanSize:=CleanSize;
      item.SpaceAfter:=(item.DataSize+item.SpaceAfter)-NewSize;
      item.DataSize:=NewSize;
      Item.Datum:=PackDate(Date);
      FStream.Position:=item.DataOffset;
      //ms.Position:=0; // not needed... CopyFrom(?,0) seeks the stream...
      FStream.CopyFrom(ms,0); // copy whole...
      //??? Clear space after:
      FillChar(buf[0],SizeOf(buf),0);
      for i:=1 to item.SpaceAfter div SizeOf(buf) do begin
        FStream.WriteBuffer(buf[0],SizeOf(buf));
      end;
      i:=item.SpaceAfter mod SizeOf(buf);
      if i>0 then FStream.WriteBuffer(buf[0],i);
      DirClean:=False;
    end else begin
      // Delete the item:
      Name:=FDir[Index];
      FDir.Delete(Index);
      SaveFreeMem(@item);
      // Append it to the archive  (but don't compress it again):
      InternalStoreItem(Name,Data,True,CleanSize,Date);
    end;
  finally ms.Free; end;
end;

procedure TPreLoadWriter.StoreItem(const Name: string; Data: TStream; const Date: TDateTime);
begin
  InternalStoreItem(Name,Data,False,0,Date);
end;

procedure TPreLoadWriter.InternalStoreItem(const Name: string; Data: TStream; IsCompressed: Boolean; CleanSize: integer; const Date: TDateTime);
var idx: integer;
    Item: PPlf_DirItem;
begin
  if not IsCompressed then begin
    idx:=IndexOfItem(Name);
    if idx>=0 then begin
      ReplaceItem(idx,Data,Date);
      exit;
    end;
  end;
  DirClean:=False;
  // Store brand-new item:
  Data.Position:=0;
  // Alloc dir-item:
  Item:=AllocMem(SizeOf(TPlf_DirItem));
  FDir.AddObject(Name,TObject(Item));
  Item.ItemSize:=SizeOf(TPlf_DirItem)-SizeOf(LongInt);
  Item.DataOffset:=FDirOffset;
  Item.Name:=Name;
  Item.Datum:=PackDate(Date);
  // Compress data:
  FStream.Position:=Item.DataOffset;
  if IsCompressed then begin
    Item.CleanSize:=CleanSize;
    FStream.CopyFrom(Data,0); // 0: copy whole...
  end else begin
    Compress(Data,FStream,Item.CleanSize);
  end;
  // Update dir-offset:
  FDirOffset:=FStream.Position;
  Item.DataSize:=FDirOffset-Item.DataOffset;
  // Align on 4-byte boundary:
  Item.SpaceAfter:=4-(FDirOffset and 3);
  if Item.SpaceAfter=4 then Item.SpaceAfter:=0;
  inc(FDirOffset,Item.SpaceAfter);
  FDirOffset:=(FDirOffset+3) and (not 3);
end;

procedure TPreLoadWriter.DeleteItem(Index: integer);
var item: PPlf_DirItem;
begin
  item:=Pointer(FDir.Objects[Index]);
  DirClean:=False;
  FDir.Delete(Index);
  if item<>nil then
    FreeMem(item);
end;

procedure TPreLoadWriter.WriteDirectory;
var Hdr: TPlf_Header;
    Dir: TPlf_DirHeader;
    i: integer;
    item: PPlf_DirItem;
begin
  // Patch header:
  FStream.Position:=0;
  FStream.ReadBuffer(Hdr,SizeOf(Hdr));
  Hdr.Version:=cPlf_Version;
  Hdr.DirOffset:=FDirOffset;
  FStream.Position:=0;
  FStream.WriteBuffer(Hdr,SizeOf(Hdr));
  // Store directory header:
  FillChar(Dir,SizeOf(Dir),0);
  Move(sPlf_Dir_Signature^,Dir.Signature[0],SizeOf(Dir.Signature));
  // Count directory size:
  Dir.NumItems:=FDir.Count;
  for i:=0 to Dir.NumItems-1 do begin
    item:=Pointer(FDir.Objects[i]);
    inc(Dir.DirSize,item.ItemSize+SizeOf(LongInt));
  end;
  Dir.Reserved:=FDirRes;
  FStream.Position:=FDirOffset;
  FStream.WriteBuffer(Dir,SizeOf(Dir));
  // Write directory items:
  for i:=0 to Dir.NumItems-1 do begin
    item:=Pointer(FDir.Objects[i]);
    FStream.WriteBuffer(item^,item.ItemSize+SizeOf(LongInt));
  end;
  SetEndOfFile(FStream.Handle);
  DirClean:=True;
end;

procedure TPreLoadWriter.ImportArchive(Src: TPreLoadObject);
var i,oidx: integer;
    item,nitem: PPlf_DirItem;
begin
  if Src=nil then exit;
  for i:=0 to Src.FDir.Count-1 do begin
    oidx:=IndexOfItem(Src.FDir[i]);
    if oidx>=0 then begin
      // duplicate item!    --> Delete current and allow import of a new one:
      item:=Pointer(FDir.Objects[oidx]);
      FDir.Delete(oidx);
      if item<>nil then FreeMem(item);
    end;
    item:=Pointer(Src.FDir.Objects[i]);
    nitem:=AllocMem(item.ItemSize+SizeOf(LongInt));
    Self.FDir.AddObject(Src.FDir[i],TObject(nitem));
    DirClean:=False;
    Move(item^,nitem^,item.ItemSize+SizeOf(LongInt));
    Src.FStream.Position:=item.DataOffset;
    nitem.DataOffset:=FDirOffset;
    Self.FStream.Position:=nitem.DataOffset;
    Self.FStream.CopyFrom(Src.FStream,nitem.DataSize);
    Self.FDirOffset:=Self.FStream.Position;
  end;
end;

{ TDbPreLoadReader }

{$ifndef NODB}

function NormalizeTableName(const TableName: string): string;
begin
  Result:=AnsiUpperCase(ChangeFileExt(ExtractFileName(TableName),''));
end;

function TDbPreLoadReader.HasTable(const TableName: string): Boolean;
begin
  Result:=(IndexOfItem(NormalizeTableName(TableName))>=0);
end;

function SkipBlank(pc: PChar): PChar;
begin
  Result:=pc;
  while Result^ in [#1..' '] do inc(Result);
end;

function SkipBlankAndSemiColon(pc: PChar): PChar;
begin
  Result:=SkipBlank(pc);
  if Result^=';' then
    Result:=SkipBlank(Result+1);
end;

const
  szNull='NULL';

// Parse date in format 'dd.mm.yyyy hh:nn:ss.zzz'
// It's not much effective, but it's late evening now...
function StrToDateFix(Value: string; var dat: TDateTime): Boolean;
var d,m,y,h,s,ms: integer;
    code: integer;
begin
  Result:=False;
  dat:=0;
  Value:=Trim(Value);
  if Pos('.',Value)<4 then begin
    // Parse date:
    System.Val(Trim(ExtractToken(Value,'.')),d,code); if code<>0 then exit;
    System.Val(Trim(ExtractToken(Value,'.')),m,code); if code<>0 then exit;
    System.Val(Trim(ExtractToken(Value,' ')),y,code); if code<>0 then exit;
    if y<100 then inc(y,2000);
    try
      dat:=EncodeDate(y,m,d);
    except {Invalid argument to date encode!} Result:=False; exit; end;
    Result:=True;
    Value:=Trim(Value);
  end;
  if Value<>'' then begin
    // Parse time:
    System.Val(Trim(ExtractToken(Value,':')),h,code); if code<>0 then exit;
    System.Val(Trim(ExtractToken(Value,':')),m,code); if code<>0 then exit;
    System.Val(Trim(ExtractToken(Value,'.')),s,code); if code<>0 then exit;
    ms:=0;
    Value:=Trim(Value);
    if Value<>'' then begin
      System.Val(Value,ms,code);
      if (code<>0) or (ms<0) or (ms>999) then ms:=0;
    end;
    // Verify values instead of catching exception...
    if (h>=0) and (h<24) and (m>=0) and (m<60) and (s>=0) and (s<60) then
      dat:=dat+EncodeTime(h,m,s,ms);
  end;
end;

procedure LoadFieldValue(Field: TField; const Value: string; WasQuoted: Boolean);
var fval: Extended;
    code: integer;
    dat: TDateTime;
begin
  if not WasQuoted and (Value=szNULL) then begin
    Field.Clear;
    exit;
  end;
  case Field.DataType of
    ftString,ftSmallint,ftInteger,ftWord
    {$ifdef VER130},ftFixedChar,ftLargeint,ftVariant {$endif VER130}: begin
      Field.AsString:=Value;
    end;
    ftFloat,ftCurrency,ftBCD: begin
      System.Val(Value,fval,code); // System.pas:_ValExt - locale-independent number-decoding (uses ".")
      if code=0 then Field.AsFloat:=fval;
    end;
    ftDate,ftTime,ftDateTime: begin
      if StrToDateFix(Value,dat) then
        Field.AsDateTime:=dat;
    end;
    ftBoolean: begin
      if Value<>'' then
        case Value[1] of
          '1'..'9': Field.AsBoolean:=True; // any number...
          '0': Field.AsBoolean:=False;
        end;
    end;
  end;
end;

procedure LoadTableLine(Tbl: TDataset; Fields: TStrings; const Line: string);
var pc,pe: PChar;
    Index,Count: Integer;
    Field: TField;
    blob: TStream;
begin
  pc:=PChar(Line);
  Tbl.Append;
  try
    Index:=0;
    Count:=Fields.Count;
    if pc<>nil then pc:=SkipBlank(pc);
    while (pc<>nil) and (pc^<>#0) do begin
      if Index>=Count then break;
      Field:=TField(Fields.Objects[Index]);
      inc(Index);
      if pc^='"' then begin
        // String constant:
        LoadFieldValue(Field,AnsiExtractQuotedStr(pc,'"'),True);
        pc:=SkipBlankAndSemiColon(pc);
      end else
      if (pc^='$') and ((pc+2)^=':') and ((pc+1)^ in ['B','b']) then begin
        // Blob:
        inc(pc,3);
        pe:=StrPos(pc,';');
        if pe=nil then pe:=StrEnd(pc);
        if (Field<>nil) and Field.IsBlob then
          blob:=Field.Dataset.CreateBlobStream(Field,bmWrite)
        else begin
          // Field is not a blob??? Dump data to some other stream instead...
          blob:=TMemoryStream.Create;
        end;
        try
          Base64DecToStream(pc^,pe-pc,blob);
        finally blob.Free; end;
        if pe^=#0 then pc:=pe
                  else pc:=SkipBlank(pe+1);
      end else begin
        LoadFieldValue(Field,pExtractToken(pc,';'),False);
        pc:=SkipBlank(pc);
      end;
    end;
    Tbl.Post;
  except
    Tbl.Cancel;
    raise;
  end;
end;

function Plf_LoadTable(Plf: TPreLoadReader; const TableName: string; Tbl: TDataset): Boolean;
var Name,FName,Line: string;
    Data: TStream;
    Size,p,len: integer;
    Srl: TStreamLineReader;
    Fields: TStringList;
    Fld: TField;
    some: Boolean;
begin
  with Plf do begin
    Name:=NormalizeTableName(TableName);
    Srl:=nil;
    Data:=nil;
    try
      if not OpenItem(IndexOfItem(Name),Data,Size) or (Size=0) then begin
        Result:=False;
        exit;
      end;
      Srl:=TStreamLineReader.Create(Data);
      Line:=Srl.ReadLine;
      Fields:=TStringList.Create;
      try
        // Parse header line:
        if Line[1]='@' then Delete(Line,1,1);
        some:=False;
        len:=Length(Line);
        p:=1;
        while (p<=len) do begin
          FName:=ExtractFieldName(Line,p);
          Fld:=Tbl.FindField(FName);
          if (Fld=nil) or (Fld.DataType=ftAutoInc) or (Fld.ReadOnly) then begin
            // Field is not known, or we wouldn't try to write to read-only fields!
            Fields.Add('');
            continue;
          end;
          some:=True;
          Fields.AddObject(FName,Fld);
        end;
        if not some then begin
          Result:=False;
          exit; // no writeable-field's-name was recognized!
        end;
        // Data lines:
        Tbl.DisableControls;
        try
          while not Srl.EOF do begin
            Line:=Srl.ReadLine;
            LoadTableLine(Tbl,Fields,Line);
          end;
          Tbl.First;
        finally Tbl.EnableControls; end;
      finally Fields.Free; end;
    finally
      Srl.Free;
      Data.Free;
    end;
    Result:=True;
  end;
end;

function TDbPreLoadReader.LoadTable(const TableName: string; Tbl: TDataset): Boolean;
begin
  Result:=Plf_LoadTable(Self,TableName,Tbl);
end;

{ TDbPreLoadWriter }

function TDbPreLoadWriter.HasTable(const TableName: string): Boolean;
begin
  Result:=(IndexOfItem(NormalizeTableName(TableName))>=0);
end;

function TDbPreLoadWriter.LoadTable(const TableName: string; Tbl: TDataset): Boolean;
begin
  Result:=Plf_LoadTable(Self,TableName,Tbl);
end;

procedure TDbPreLoadWriter.StoreTable(const TableName: string; Tbl: TDataset);
var Fields: TStringList;
    i: integer;
    Fld: TField;
begin
  Tbl.Active:=True;
  Fields:=TStringList.Create;
  try
    // Get list of fields:
    for i:=0 to Tbl.FieldCount-1 do begin
      Fld:=Tbl.Fields[i];
      // Don't store read-only fields:
      if (Fld.DataType=ftAutoInc) {or (Fld.ReadOnly and not Tbl.ReadOnly){} then
        continue;
      Fields.AddObject(Fld.FieldName,Fld);
    end;
    if Fields.Count=0 then
      exit;
    StoreTableEx(TableName,Tbl,Fields);
  finally
    Fields.Free;
  end;
end;

function FloatToStrFix(const Value: Extended): string;
var p,len: integer;
begin    
  // Uses System.pas:_Str2Ext ... It produces result invariant to windows locale...
  Str(Value:1:6,Result);
  p:=Pos('.',Result);
  if p>0 then begin
    // Eat terminating zeros:
    len:=Length(Result);
    if Result[len]='0' then begin
      dec(len);
      while (len>p) and (Result[len]='0') do dec(len);
      inc(len); // keep only last 0...
      SetLength(Result,len);
    end;
  end;
end;

procedure SpacesToZero(var S: string);
var p: integer;
    pc: PChar;
begin
  UniqueString(S); pc:=Pointer(S); // Delphi 5 could write: pc:=@S[1]; // Maybe other versions too???
  for p:=1 to Length(S) do
    if pc^=' ' then
      pc^:='0'; // It's faster to write to a pchar, because S[p]:='0' would call UniqueString for each write!
end;

// This would return similar to FormatDateTime('dd.mm.yyyy hh:nn:ss.zzz',Date);
// but invariant to user-selected windows locale...

function DateToStrFix(const Date: TDateTime; FType: TFieldType): string;
var sd,st: string;
    y,m,d,h,s,ms: word;
begin
  sd:='';
  st:='';
  if FType in [ftDate,ftDateTime] then begin
    DecodeDate(Date,y,m,d);
    sd:=Format('%2d.%2d.%4d',[d,m,y]);
    SpacesToZero(sd); // It would be nice, if format understood '%02d' ...
  end;
  if FType in [ftTime,ftDateTime] then begin
    DecodeTime(Date,h,m,s,ms);
    if ms=0 then
      st:=Format('%2d:%2d:%2d',[h,m,s])
    else
      st:=Format('%2d:%2d:%2d.%3d',[h,m,s,ms]);
    SpacesToZero(sd); // It would be nice, if format understood '%02d' ...
  end;
  Result:=Trim(sd+' '+st);
end;

function CollectList(List: TStrings; const Separator,Prefix: string): string; forward;

function DumpField(Field: TField): string;
var blob: TStream;
    enc: TStringList;
    IsNull: Boolean;
Label _Blob,_Null;
begin
  if Field.IsBlob then goto _Blob; // Don't test IsNull for a blob in vain!
  if Field.IsNull then begin
  _Null:
    Result:=szNULL;
    exit;
  end;
  case Field.DataType of
    ftString{$ifdef VER130},ftFixedChar {$endif VER130}:
      Result:=AnsiQuotedStr(Field.AsString,'"');
    ftSmallint,ftInteger,ftWord,ftAutoInc
    {$ifdef VER130},ftLargeint,ftVariant {$endif VER130}:
      Result:=Field.AsString;
    ftFloat,ftCurrency,ftBCD:
      Result:=FloatToStrFix(Field.AsFloat);
    ftDate,ftTime,ftDateTime:
      Result:=DateToStrFix(Field.AsDateTime,Field.DataType);
    ftBoolean:
      Result:=IntToStr(ord(Field.AsBoolean));
    //ftWideString: Result:=''; // Cannot be stored into AnsiString result...
    //
    //ftBytes,ftVarBytes,ftBlob,ftMemo,ftFmtMemo,ftGraphic,ftTypedBinary: begin
    else begin
      if Field.IsBlob then begin
      _Blob:
        IsNull:=False;
        blob:=Field.Dataset.CreateBlobStream(Field,bmRead);
        try
          if (blob.Size=0) then begin
            IsNull:=True;
          end else begin
            enc:=TStringList.Create;
            try
              Base64EncStreamToLines(blob,enc);
              Result:=CollectList(enc,'','$B:');
            finally enc.Free; end;
          end;
        finally blob.Free; end;
        if IsNull then goto _Null; // Couldn't goto from middle of try..finally
      end else
        Result:=''; // unknown field-type...
    end;
  end;
end;

function CollectList(List: TStrings; const Separator,Prefix: string): string;
var i,len,rlen,Count,lsep,lprefix: integer;
    pres: PChar;
    Value: string;
begin
  // This could be little optimized by caching results of Length(List[?])...
  // But it would require more memory or stack space...
  Count:=List.Count;
  lsep:=Length(Separator);
  lprefix:=Length(Prefix);
  rlen:=lprefix+Count*lsep; // all semi-colons...
  for i:=0 to Count-1 do
    inc(rlen,Length(List[i]));
  SetLength(Result,rlen);
  // Next line could be also written:   pres:=@Result[1]; // Delphi 5 would call UniqueString...
  UniqueString(Result); pres:=Pointer(Result);
  // Prefix:
  if lprefix>0 then begin
    Move(Pointer(Prefix)^,pres^,lprefix);
    inc(pres,lprefix);
  end;
  // Lines:
  for i:=0 to Count-1 do begin
    Value:=List[i];
    len:=Length(Value);
    Move(Pointer(Value)^,pres^,len);
    inc(pres,len);
    if lsep>0 then begin
      Move(Pointer(Separator)^,pres^,lsep);
      inc(pres,lsep);
    end;
  end;
  SetLength(Result,rlen-lsep); // eat last ";" ...
end;

function DumpTableLine(Tbl: TDataset; Fields,FieldValues: TStrings): string;
var i,Count: integer;
begin
  FieldValues.Clear;
  Count:=Fields.Count;
  for i:=0 to Count-1 do begin
    FieldValues.Add(DumpField(TField(Fields.Objects[i])));
  end;
  Result:=CollectList(FieldValues,';','');
end;

procedure TDbPreLoadWriter.StoreTableEx(const TableName: string; Tbl: TDataset; Fields: TStrings);
var i: integer;
    Bmk,Line: string;
    FVals: TStringList;
    Data: TStream;

  procedure WriteLine;
  begin
    Line:=Line+#13#10;
    Data.WriteBuffer(Line[1],Length(Line));
  end;

begin
  if (Fields=nil) or (Fields.Count=0) then begin
    // Let normal StoreTable fill the fields:
    StoreTable(TableName,Tbl);
    exit;
  end;
  Tbl.Active:=True;
  //
  FVals:=TStringList.Create;
  Data:=TMemoryStream.Create;
  try
    // Header line (field-names):
    Line:='@';
    for i:=0 to Fields.Count-1 do 
      Line:=Line+Fields[i]+';';
    SetLength(Line,Length(Line)-1);
    WriteLine;
    // Data lines:
    Bmk:=Tbl.Bookmark;
    Tbl.DisableControls;
    try
      Tbl.First;
      while not Tbl.EOF do begin
        Line:=DumpTableLine(Tbl,Fields,FVals);
        WriteLine;
        Tbl.Next;
      end;
    finally
      try Tbl.Bookmark:=Bmk; except ; end;
      Tbl.EnableControls;
    end;
    // Store Data:
    StoreItem(NormalizeTableName(TableName),Data,SysUtils.Date);    
  finally
    Data.Free;
    FVals.Free;
  end;
end;

{$endif NOT NODB}

end.

